/* ------------------------------------------------------------------------
 * $Id: client.cc,v 1.21 2001/12/07 13:09:34 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2000-09-19 by Steve Houston.
 *
 * Copyright (c) 2000, 2001 Niklas Elmqvist <elm@3dwm.org>.
 * Copyright (c) 2000, 2001 Steve Houston <steve@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System Includes
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <time.h>

// -- 3Dwm Includes
#include <Celsius/GetOpt.hh>
#include <Nobel++/Nobel.hh>
#include <Nobel/Shape.hh>
#include <Nobel/Image.hh>
#include <Nobel/Mapper.hh>
#include <Nobel/Desktop.hh>
#include <Nobel/NodeKit.hh>
#include <Nobel/Appearance.hh>
#include <Nobel/GeometryKit.hh>
#include <Nobel/TextureKit.hh>
#include <Nobel/TriangleGeometry.hh>
#include <Nobel/TextureGenerator.hh>

using namespace Nobel;

// -- Defines
#define VERSION		"0.2"

// -- Class Definitions

class VNCClient {
public:

    /**
     * Constructor.
     **/ 
    VNCClient(int &argc, char **argv, const GetOpt &opt)
	: _client("VNC Client", argc, argv),
	  _opt(opt) {
	
	// Resolve the necessary kits
	_node_kit = _client.resolve<NodeKit>(Nobelxx::name<NodeKit>());
	_geo_kit = _client.resolve<GeometryKit>(Nobelxx::name<GeometryKit>());
	_tex_kit = _client.resolve<TextureKit>(Nobelxx::name<TextureKit>());
	
	// Create the VNC session
	createSession(Node_var(_client.root()));
    }

private: 
    
    bool isnumerical(char c) {
	return isdigit(c) || c == '.' || c == '-';
    }
    
    std::string parseValue(const std::string &str, float &value, char delim) {
	std::string::size_type begin = 0;
	while (begin < str.length() && str[begin] == delim)
	    begin++;
	std::string::size_type end = begin;
	while (end < str.length() && isnumerical(str[end]))
	    end++;
	if (begin == end) return str.substr(begin);
	std::string number = str.substr(begin, end);
	value = strtod(number.c_str(), NULL);
	return str.substr(end);
    }

    /**
     * Application geometry. 
     **/ 
    struct Geometry {
	float width, height, depth;
	float x, y, z;
	float xrot, yrot, zrot;
	Geometry(float _width = 0, float _height = 0, float _depth = 0,
		 float _x = 0, float _y = 0, float _z = 0,
		 float _xrot = 0, float _yrot = 0, float _zrot = 0) 
	    : width(_width), height(_height), depth(_depth),
	      x(_x), y(_y), z(_z),
	      xrot(_xrot), yrot(_yrot), zrot(_zrot)
	{ }
    };
    
    void parseGeometry(const std::string &s, Geometry &geo) {
	std::string str = s;

	// Parse position
	str = parseValue(str, geo.width, 'x');
	str = parseValue(str, geo.height, 'x');
	str = parseValue(str, geo.depth, 'x');
	
	// Parse dimensions
	str = parseValue(str, geo.x, '+');
	str = parseValue(str, geo.y, '+');
	str = parseValue(str, geo.z, '+');

	// Parse rotation
	str = parseValue(str, geo.xrot, '*');
	str = parseValue(str, geo.yrot, '*');
	str = parseValue(str, geo.zrot, '*');
    }
    
    void createSession(Node_ptr root) {

	// Retrieve session parameters
	const char *host = _opt.getArgument(0).c_str();
	int port = atoi(_opt.getArgument(1).c_str());
	const char *passwd = _opt.getArgument(2).c_str();
	
	// Retrieve window geometry
	Geometry geo(10, 10, 1);
	std::string geo_str = _opt.get("geometry");
	if (geo_str != "") 
	    parseGeometry(geo_str, geo);
	
	// Create the triangle shape for the plane
	Shape_var shape = _node_kit->createShape();
	TriangleGeometry_var tri = _geo_kit->createTriangleGeometry();
	
	// Connect to the VNC desktop
	Desktop_var desktop = _node_kit->createVNCDesktop(host, port, passwd);
	DesktopController_var controller = desktop->createController(shape);

	// Create a planar mapper to use in the controller
	Mapper_var mapper = _tex_kit->createPlanarMapper();
	mapper->scale(Nobelxx::Vertex3D(1 / geo.width * 2,
					1 / geo.height * 2, 1.0));
	controller->setMapper(mapper);
	
	// Retrieve and link the texture generator to our procedural texture
	TextureGenerator_var texGen = desktop->getTextureGenerator();
	ProceduralTexture_var tex = _tex_kit->createProceduralTexture();
	tex->setTextureGenerator(texGen);
	
	// Link this texture to the appearance object
	Appearance_var appear = _geo_kit->createAppearance();
	appear->setTexture(tex);
	
	// Set the appearance into the shape
	shape->setAppearance(appear);
	
	// Initialise the vertex list
	Vertex3D vertex;
	tri->setVertexNumber(4);
	
	vertex.x = -0.5f * geo.width + geo.x;
	vertex.y = -0.5f * geo.height + geo.y;
	vertex.z =  geo.z;
	tri->setVertex(0, vertex);
	
	vertex.x =  0.5f * geo.width + geo.x;
	vertex.y = -0.5f * geo.height + geo.y;
	vertex.z =  geo.z;
	tri->setVertex(1, vertex);
	
	vertex.x =  0.5f * geo.width + geo.x;
	vertex.y =  0.5f * geo.height + geo.y;
	vertex.z =  geo.z;
	tri->setVertex(2, vertex);
	
	vertex.x = -0.5f * geo.width + geo.x;
	vertex.y =  0.5f * geo.height + geo.y;
	vertex.z =  geo.z;
	tri->setVertex(3, vertex);
	
	// Initialise the tex coordinate list
	TexCoord tc;
	tri->setTexCoordNumber(4);
	
	tc.u = 0.0f;
	tc.v = 1.0f;
	tri->setTexCoord(0, tc);
	
	tc.u = 1.0f;
	tc.v = 1.0f;
	tri->setTexCoord(1, tc);
	
	tc.u = 1.0f;
	tc.v = 0.0f;
	tri->setTexCoord(2, tc);
	
	tc.u = 0.0f;
	tc.v = 0.0f;
	tri->setTexCoord(3, tc);
	
	// Set the indices
	tri->setFaceNumber(2);
	Triangle face, texCoordFace;
	
	face.a = 0;
	face.b = 1;
	face.c = 3;
	texCoordFace.a = 0;
	texCoordFace.b = 1;
	texCoordFace.c = 3;
	tri->setVertexIndex(0, face);
	tri->setTexCoordIndex(0, texCoordFace);
	
	face.a = 1;
	face.b = 2;
	face.c = 3;
	texCoordFace.a = 1;
	texCoordFace.b = 2;
	texCoordFace.c = 3;
	tri->setVertexIndex(1, face);
	tri->setTexCoordIndex(1, texCoordFace);
	
	// Add geometry to shape and shape to root
	shape->addGeometry(tri);
	root->insert(controller);	
    }
    
private:
    Nobelxx::Client _client;
    GetOpt _opt;    
    NodeKit_var _node_kit;
    GeometryKit_var _geo_kit;
    TextureKit_var _tex_kit;
};

// -- Code Segment

int main(int argc, char *argv[]) {
    
    try {

	// Create a command-line parser
	GetOpt opt("3Dwm VNC client", VERSION, 
		   "[options] <server> <port> <password>");
	
	// Add some command-line options
	opt.add('g', "geometry", GetOpt::mandatory,
		"application geometry (WxHxD+X+Y+Z*Xr*Yr*Zr)");

	// Parse the command-line
	if (opt.parse(argc, argv) == false)
	    return EXIT_FAILURE;
	
	// Make sure we got enough arguments
	if (opt.getArgumentCount() < 3) {
	    std::cerr << argv[0] << ": mandatory arguments missing!"
		      << std::endl;
	    opt.usage(argv[0]);
	    return EXIT_FAILURE;
	}
	
	// Create a client instance
	VNCClient client(argc, argv, opt);

	// Do stuff
    }
    catch (const CORBA::Exception &e) {

	// The ORB threw an exception
	std::cerr << "Uncaught CORBA exception: " << e << std::endl;
	return EXIT_FAILURE;
    }
    catch (...) {
	return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}
